#ifndef LIB_BVHFILE_H
#define LIB_BVHFILE_H

#include <vector>
#include <string>
#include <assert.h>

#include "math/vec.h"
#include "base/view.h"
#include "base/texture.h"

#include "lib/lexeme.h"

enum {
	BVH_CHANNEL_XPOS,
	BVH_CHANNEL_YPOS,
	BVH_CHANNEL_ZPOS,
	BVH_CHANNEL_XROT,
	BVH_CHANNEL_YROT,
	BVH_CHANNEL_ZROT
};

enum {
	BVH_NODETYPE_ROOT=1,
	BVH_NODETYPE_JOINT,
	BVH_NODETYPE_ENDSITE
};

//#define BVH_DEFORMABLE_MESH
//#define BVH_STORE_HISTORY
//#define BVH_STORE_HISTORY_COUNT		100		//number of records to hold
//#define BVH_DEFORMABLE_MESH_HISTORY		//uses the history buffer to store the mesh rather than the skeleton timestep mesh

class BVHNode {
public:
	vec3f offset;

	//index in the BVH file
	//used to access the basis array's
	int nodeIndex;

	//index of its parent in the nodes[] vector
	int parentIndex;

	//// per-frame dynamic variables:

	//world basis based upon its' channel's basis info and its parent's worldbasis
	basis34f basis;

	//// methods:

	virtual int getType() = 0;

	typedef struct {
		int first;
		int count;
	} vtx_ring_t;

	typedef struct {
		vec3f	x;	//point of the vertex
		int		n;	//index of the node associated with this ring
		float	t;	//coefficient of interpolation from the parent to the current node's basis
	} vtx_t;

	virtual vtx_ring_t genVertexRing(std::vector<vtx_t> &vtxList, int nodeIndex, float coeff, const vec3f &normal);
	virtual void bindVertexRing(std::vector<int> &triList, const vtx_ring_t &r1, const vtx_ring_t &r2);

	virtual vtx_ring_t genRing(std::vector<BVHNode *> &nodeList, std::vector<int> &triList, std::vector<vtx_t> &vtxList);
};

class BVHGroup : public BVHNode {
public:
	std::string name;

	//currently these are only used during the MOTION data parsing.  never during realtime.
	int channelIndex;	//index within the channels[] vector as to where the first of this node is located
	int channelCount;	//number of channels for this object

	std::vector<BVHNode *> child;

	BVHGroup() {
		channelIndex = 0;
		channelCount = 0;
	}

#if 0
	void render(
		const int *channelFrame,
		const basis34f *basisFrame,
		const basis34f *nextBasisFrame,
		float coeff);
#endif
};

class BVHRoot : public BVHGroup {
public:
	virtual int getType() { return BVH_NODETYPE_ROOT; }
};

class BVHJoint : public BVHGroup {
public:
	virtual int getType() { return BVH_NODETYPE_JOINT; }
};

class BVHEndSite : public BVHNode {
public:
	virtual int getType() { return BVH_NODETYPE_ENDSITE; }
};

class BVHFile {
protected:

	//instance vars

	std::vector<BVHNode *> nodes;

	//the fraction of second per each frame (one over framerate)
	float frameFreq;

	std::vector<int> channels;	//maps the index to a channeltype: tx,ty,tz,rx,ry,rz

	//the number of frames for this bvh's data
	int frameCount;

	//the current frame
	float frame;

	//this will be sized 'frameCount' by the node count
	basis34f *basis;

	//load methods
		//generic
	void load_error(LexemeParser *lex, const char *_format, ...);
	void load_expected(LexemeParser *lex, const char *str);
	bool load_matches(LexemeParser *lex, const char *str, bool must = true);
		//variable
	bool load_read_int(LexemeParser *lex, const char *name, int *dest);
	bool load_read_float(LexemeParser *lex, const char *name, float *dest);
	bool load_read_string(LexemeParser *lex, const char *name, char *dest, int size);
		//specific
	bool load_motion(LexemeParser *lex);
	BVHNode *load_node(LexemeParser *lex, int type, int parentIndex);
	bool load_heirarchy(LexemeParser *lex);

	void reset() {
		frameCount = 0;
		frameFreq = 0;
		frame = 0;
		basis = NULL;
#ifdef BVH_STORE_HISTORY
		history = NULL;
		history_freq = (float)BVH_STORE_HISTORY_COUNT;
		history_index = 0;
#endif
#ifdef BVH_DEFORMABLE_MESH
		curVtxSet = NULL;
		curNrmSet = NULL;
#endif
	}

#ifdef BVH_STORE_HISTORY
	int history_index;		//frame within the buffer
	float history_freq;		//add a new image to the buffer when the frame counter passes over this
	//sized nodes.size() by BVH_STORE_HISTORY_COUNT
	//todo - unify this basis set with the node's basis set?
	basis34f *history;

	void storeHistory(float dt);
#endif //BVH_STORE_HISTORY

#ifdef BVH_DEFORMABLE_MESH
	//deformable mesh:
	std::vector<int> triList;
	std::vector<BVHNode::vtx_t> vtxList;
	vec3f *curVtxSet;
	vec3f *curNrmSet;

	void calcDeformMesh();
#endif //BVH_DEFORMABLE_MESH

public:

	const Texture *skyTex;

	BVHFile() {reset();}
	~BVHFile() {shutdown();}

	bool init(const char *filename);
	void shutdown();

	//sets the current frame
	void setFrame(float frame);

	//generates the basii for the current specified 'frame' value
	void genBasis();
	void genIdentityBasis();

	//wrapper for setFrame(frame + df(dt)); genBasis();
	void update(float dt);

	//renders the bvh according to the current set of basii's
	void render(const View *view);

	const std::vector<BVHNode *> &getNodes() const { return nodes; }
	const std::vector<int> &getChannels() const { return channels; }

	float getFrameFreq() { return frameFreq; }

};

#endif

/*
Here's our issue:

for each node we have:
	translation
		base through OFFSET field
		possible animation through XYZposition CHANNELs
	rotation
		base identity
		possible animation through XYZrotation CHANNELs

here's what I'm thinking:
since our MOTION data is broken into frames, and each frame's channel data aligns with the HEIRARCHY structure :

convert the MOTION data into pos/rot data

bone0:
tx
ty			=> trans
tz
rz
rx			=> rot
ry
bone1:
rz
rx			=> rot
ry
bone2:
rz
rx			=> rot
ry

so
	- for every frame
		- for every bvh skel node
			- make us a pos/rot vec3f/vec4f (basis34f-style)
			- and interpolate between on the render's

*/
