#ifndef LIB_OBJFILE_H
#define LIB_OBJFILE_H

#include <vector>

#include "defs/defs.h"		//VECTOR_FIX_MSVC50
#include "math/vec.h"
#include "base/texture.h"

/**
 * the data associated with each index
 */
typedef struct {

	/** 
	 * reference to the vertex array
	 */
	int v;

	/** 
	 * reference to the texture coordinate array
	 */
	int t;

	/** 
	 * reference to the normal array
	 */
	int n;

} triIndex_t;

/**
 * information associated with each triangle
 */
typedef struct {

	/**
	 * three vertex references
	 */
	triIndex_t	p[3];

	/** 
	 * surface plane equation
	 */
	plane3f		plane;

	/**
	 * group associated with this material
	 */
	int		mtlGroupIndex;

} tri_t;

/**
 * triangle connectivity information
 */
typedef struct {
	int t[2];		//two triangles between the edge
	int v[2];		//the first vertices of each triangle
					//triangle t[0] will have v[0]->(v[0]+1) in the correct right hand direction
					//triangle t[1] will be backwards: (v[0]+1)->v[0]
} edge_t;

class ObjFileMaterial {
public:

	//default:
	ObjFileMaterial() {
		illum = 0;
		Ns = 1;
	}

	//specified:
	std::string name;
	vec3f Ka;				//ambient
	vec3f Kd;				//diffuse
	vec3f Ks;				//specular
	int illum;				//0 = none, 1 = diffuse, 2 = specular
	float Ns;				//shininess
	std::string map_Kd;		//filename to diffuse map

	//derived:
	Texture map_Kd_tex;
};

class ObjFile {
public:

	enum {
		// load flags
		GEN_EDGES,
		REGEN_NORMALS,
		// state flags.  i should make these ignore_*** and make them double as load flags too ... hmm ...
		USE_TEXCOORDS,
		USE_NORMALS,
	};

protected:

	/**
	 * bitflags for bits indexed by OBJFILE_FLAG_***
	 * holds specified attributes about this object
	 */
	int flags;

	/**
	 * opengl compiled mode
	 */
	GLuint	list;
	
	/**
	 * opengl vertex array mode
	 */
	GLuint *indexArrayBuffer;

	/**
	 * identifier -- typically the filename
	 */
	char title[256];

	/**
	 * an array of the vertex coordinates
	 */
	std::vector<vec3f> vertex;

	/**
	 * an array of the normals, sized by vertexCount
	 */
	std::vector<vec3f> normal;

	/**
	 * an array of the texcoords, sized by vertexCount
	 */
	std::vector<vec2f> texCoord;

	//todo - for optimization with glDrawElements,
	//store one vector of unique <vertex/texcoord/normal> sets

	/**
	 * an array of the <vertex/texcoord/normal> index sets for each triangle vertex
	 */
	std::vector<tri_t> tri;

	/**
	 * array of pairs of integers referencing triangle indices
	 * ranged [0, 2 * tri.size() / 3 - 1]
	 */
	std::vector<edge_t> edge;

	/**
	 * Bounding box containing all the vertices of this object
	 */
	box3f bbox;

	/**
	 * Bounding radius containing all the vertices of this object
	 */
	float brad;


	/**
	 * Clear the object
	 */
	void reset();

	/**
	 * loads the specified material file and stores it with this model
	 */
	bool loadMaterialFile(const char *filename);

	/**
	 * Materials associated with this object file
	 */
	std::vector<ObjFileMaterial *> materials;

public:

	ObjFile() {reset();}
	~ObjFile() {unload();}

	/**
	 * Unallocate all buffers and clear the object
	 */
	void unload();

	/**
	 * Loads the ObjFile object from a file
	 */
	bool load(const char *filename, int loadflags = bitflag(GEN_EDGES));

	/**
	 * Renders the object in immediate mode.  Uses the specified vertex array in place of its own
	 */
	void renderImmediate(const vec3f *vtxs, bool useMtl = true) const;

	void renderImmediate(bool useMtl = true) const { renderImmediate(&vertex[0], useMtl); }
	
	void renderImmediateGroupOnly(GLuint group);

	/**
	 * Render an object using normals
	 */
	void render() const;
	
	/**
	 * Render with vertex arrays enabled for vertex, normal, and texcoord
	 */
	void renderWithArrays();

	/**
	 * Render an object using normals, extend along normal direction.  immediate mode only.
	 */
	void renderExtendNormals(float magnitude);

	/**
	 * Compiles the model into retained mode.
	 */
	void compile();

	/**
	 * Deletes display list
	 */
	void uncompile();

	/**
	 * whether the model is compiled or not
	 */
	bool isCompiled() { return list!=0; }

	/**
	 * constant access to the bounding box
	 */
	const box3f &getBBox() const { return bbox; }

	/**
	 * returns the radius of the smallest sphere centered at the origin 
	 * surrounding all vertices of this object
	 */
	const float getBoundingRadius() const { return brad; }

	/**
	 * access to the title
	 */
	const char *getTitle() { return title; }

	/**
	 * vertex array access
	 */
	const std::vector<vec3f> &getVertexes() const { return vertex; }

	/**
	 * normal array access
	 */
	const std::vector<vec3f> &getNormals() const { return normal; }

	/**
	 * texture coordinate array access
	 */
	const std::vector<vec2f> &getTexCoords() const { return texCoord; }

	/**
	 * triangle array access
	 */
	const std::vector<tri_t> &getTris() const { return tri; }

	/** 
	 * edge array access
	 */
	const std::vector<edge_t> &getEdges() const { return edge; }

	/**
	 * access to the flags for the model:
	 */
	int getFlags() { return flags; }
};

VECTOR_FIX_MSVC50(tri_t)
VECTOR_FIX_MSVC50(edge_t)

#endif
